﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Apache.NMS.ActiveMQ;
using Apache.NMS;
using System.Threading;

namespace ActiveMQ_Test
{
public class ActiveMQClient
{
    private string uri, destinationQueue;
    private IMessageProducer producer;
    private Dictionary<string, AsyncMessageHelper> responseBuffer;
    private ITemporaryQueue temporaryQueue;
    private ISession session;
    private IConnection connection;

    public ActiveMQClient(string uri, string destinationQueue)
    {
        this.uri = uri;
        this.destinationQueue = destinationQueue;
        this.responseBuffer = new Dictionary<string, AsyncMessageHelper>();

        // Create the connection factory which will be used to create new connections to the queue.  We set the URI to that of the ActiveMQ server.
        var connectionFactory = new ConnectionFactory(this.uri);

        try
        {
            // Create and open a new connection to ActiveMQ.
            connection = connectionFactory.CreateConnection();
            connection.Start();

            // Create a new session.
            this.session = connection.CreateSession(AcknowledgementMode.AutoAcknowledge);

            // Get a handle to the destination queue by its queue name.
            var destination = session.GetDestination(destinationQueue);

            // Setup a message producer to send message to the queue the server is consuming from.
            this.producer = session.CreateProducer(destination);
            this.producer.DeliveryMode = MsgDeliveryMode.NonPersistent;

            // Create a temporary queue for this session.  This queue will survive until the connection is closed, and then it
            // is automatically deleted by ActiveMQ.  This temporary queue will be used by the consumer (server) to send response
            // messages to.
            this.temporaryQueue = session.CreateTemporaryQueue();

            // Create a consumer to listen for messages on the temporary queue we just created.
            var responseConsumer = session.CreateConsumer(temporaryQueue);
            // Wire up an event which will be fired upon receipt of response messages.
            responseConsumer.Listener += new MessageListener(responseConsumer_Listener);
        }
        catch (Exception ex)
        {
            Console.WriteLine(ex);
        }
    }

    /// <summary>
    /// Disposes of the ActiveMQ connection.
    /// </summary>
    public void Stop()
    {
        try
        {
            this.connection.Dispose();
        }
        catch (Exception)
        {
        }
    }

    /// <summary>
    /// Sends a message and blocks until a response is received or the timeout expires.
    /// </summary>
    /// <param name="message"></param>
    /// <returns></returns>
    public IMessage SendMessage(IMessage message, int timeout = 10000)
    {
        // Create a unique correlation ID which we will use to map response messages to request messages.
        var correlationID = Guid.NewGuid().ToString();
        message.NMSCorrelationID = correlationID;

        // Set the reply-to header to the temporary queue that we created so that the server knows where to return messages.
        message.NMSReplyTo = this.temporaryQueue;

        // Create a new AsyncMessageHelper.  This class is a helper class to make it easier for us to map response messages to request messages.
        using (var asyncMessageHelper = new AsyncMessageHelper())
        {
            // Add the async helper to the response buffer.
            lock (this.responseBuffer)
                this.responseBuffer[correlationID] = asyncMessageHelper;

            // Send the message to the queue.
            this.producer.Send(message);

            // Wait for a response for up to [timeout] seconds.  This blocks until the timeout expires or a message is received (.Set() is called on the trigger then, allowing execution to continue).
            asyncMessageHelper.Trigger.WaitOne(timeout, true);

            // Either the timeout has expired, or a message was received with the same correlation ID as the request message.
            IMessage responseMessage;
            try
            {
                // The Message property on the async helper will not have been set if no message was received within the timeout period.
                if (asyncMessageHelper.Message == null)
                    throw new TimeoutException("Timed out while waiting for a response.");

                // We got the response message, cool!
                responseMessage = asyncMessageHelper.Message;
            }
            finally
            {
                // Remove the async helper from the response buffer.
                lock (this.responseBuffer)
                    this.responseBuffer.Remove(correlationID);
            }

            // Return the response message.
            return responseMessage;
        }
    }

    /// <summary>
    /// Creates a text message for use with the client.
    /// </summary>
    public ITextMessage CreateTextMessage(string text = null)
    {
        return text == null
            ? this.session.CreateTextMessage()
            : this.session.CreateTextMessage(text);
    }

    #region Event Handlers

    /// <summary>
    /// This event will fire when a response message is received from the temporary queue.  It will attempt to map received messages back
    /// to the response buffer.
    /// </summary>
    void responseConsumer_Listener(IMessage message)
    {
        // Look for an async helper with the same correlation ID.
        AsyncMessageHelper asyncMessageHelper;
        lock (this.responseBuffer)
        {
            // If no async helper with the same correlation ID exists, then we've received some erranious message that we don't care about.
            if (!this.responseBuffer.TryGetValue(message.NMSCorrelationID, out asyncMessageHelper))
                return;
        }

        // Set the Message property so we can access it in the send method.
        asyncMessageHelper.Message = message;

        // Fire the trigger so that the send method stops blocking and continues on its way.
        asyncMessageHelper.Trigger.Set();
    }

    #endregion

    #region Nested Classes

    /// <summary>
    /// Helper class which assists with keeping track of message timeouts, blocking the sending thread and carrying messages through the async eventing back
    /// to the synchronous call.
    /// </summary>
    private class AsyncMessageHelper : IDisposable
    {
        public IMessage Message { get; set; }
        public AutoResetEvent Trigger { get; private set; }

        public AsyncMessageHelper()
        {
            this.Trigger = new AutoResetEvent(false);
        }

        #region IDisposable Members

        public void Dispose()
        {
            this.Trigger.Dispose();
            this.Message = null;
        }

        #endregion
    }

    #endregion
}
}
