﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Apache.NMS;
using Apache.NMS.ActiveMQ;

namespace ActiveMQServer
{
    public class ActiveMQServer
    {
        private string uri, destinationQueue;
        private ISession session;
        private IMessageProducer replyProducer;
        private IConnection connection;

        public ActiveMQServer(string uri, string destinationQueue)
        {
            this.uri = uri;
            this.destinationQueue = destinationQueue;

            // Create the connection factory which will be used to create new connections to the queue.  We set the URI to that of the ActiveMQ server.
            var connectionFactory = new ConnectionFactory(uri);

            // Create and open a new connection to ActiveMQ.
            connection = connectionFactory.CreateConnection();
            connection.Start();

            // Create a new session.
            this.session = connection.CreateSession(AcknowledgementMode.AutoAcknowledge);

            // Get a handle to the destination queue by its queue name.
            var queue = this.session.GetDestination(destinationQueue);

            // Create and setup a producer to send response messages back to the client.  Note that we do not
            // set a destination here because it will be provided with each individual reply.
            this.replyProducer = this.session.CreateProducer();
            this.replyProducer.DeliveryMode = MsgDeliveryMode.NonPersistent;

            // Create and setup a consumer to consume (receive) messages from the queue that the client is sending messages to.
            var consumer = this.session.CreateConsumer(queue);
            // Wire-up an event to be fired when a message is received from the queue.
            consumer.Listener += new MessageListener(consumer_Listener);
        }

        /// <summary>
        /// Disposes of the ActiveMQ connection.
        /// </summary>
        public void Stop()
        {
            try
            {
                this.connection.Dispose();
            }
            catch (Exception)
            {
            }
        }

        /// <summary>
        /// This method is fired upon receipt of a message from the client queue.  It will print the received message to the console and then
        /// send a response message back to the client.
        /// </summary>
        void consumer_Listener(IMessage message)
        {
            try
            {
                // Create the response message.  We'll send a simple text-based message back.
                var response = this.session.CreateTextMessage();

                // Determine the text to send back to the client.
                var textMessage = message as ITextMessage;
                if (textMessage == null)
                    response.Text = "Expected to receive an ITextMessage...";
                else
                {
                    Console.WriteLine("Received: " + textMessage.Text);
                    response.Text = "ECHO(" + textMessage.Text + ")";
                }

                // Set the correlation ID to that of the received message.
                response.NMSCorrelationID = message.NMSCorrelationID;

                // Send the response message to the reply-to destination as received in the message header.  This is the
                // temporary queue that we created in the client application.
                this.replyProducer.Send(message.NMSReplyTo, response);
            }
            catch (NMSException ex)
            {
                Console.WriteLine("Error: {0}", ex);
            }
        }
    }
}
